"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = require("axios");
const log = require("fary-log");
const logger = log('fary:proxy-load-balancer');
const MATH = Math;
/**
 * Load balance for proxy
 * Currently use random algorithm.
 * At the beginning, considering all the targets are healthy.
 * Check every proxy target if they are healthy intervarlly.
 * If all the targets are not healthy, it will throw error when method `pick` is invoked.
 */
class ProxyLoadBalancer {
    /**
     * constructor
     * @param {IProxyTargets} targets - proxy targets
     * @param {IProxyLoadBalancerOptions} options - options
     */
    constructor(targets, { enableHealthCheck = false, healthCheckPath = '/admin/health', healthCheckTimeout = 1000, healthCheckInterval = 1000 * 60, } = {}) {
        if (typeof targets === 'string') {
            targets = [targets];
        }
        if (!Array.isArray(targets)) {
            throw Error(`Proxy target must be string or array, got '${typeof targets}'`);
        }
        this.totalTargets = targets;
        this.healthyTargets = targets;
        this.options = {
            enableHealthCheck,
            healthCheckPath,
            healthCheckTimeout,
            healthCheckInterval,
        };
        if (this.options.enableHealthCheck) {
            this.startIntervalCheck();
        }
    }
    /**
     * setup health check interval timer
     */
    startIntervalCheck() {
        setInterval(() => {
            this.checkHealth();
        }, this.options.healthCheckInterval);
    }
    /**
     * the main process of health check
     */
    checkHealth() {
        logger.log('Health check starting...');
        const checkPromises = this.totalTargets.map((target) => {
            const checkUrl = `${target}${this.options.healthCheckPath}`;
            logger.log(`Health check url: '${checkUrl}'`);
            return axios_1.default.get(checkUrl, {
                timeout: this.options.healthCheckTimeout,
                maxRedirects: 0,
                validateStatus: (status) => {
                    return status === 200;
                },
            })
                .then(() => {
                logger.log(`Health check passed for '${target}'`);
                return target;
            })
                .catch((error) => {
                logger.error(`Health check failed for '${target}', reason:`);
                logger.error(error);
                return '';
            });
        });
        return Promise.all(checkPromises)
            .then((results) => {
            // update the healthy targets list, removing  the target which does not return 200 status
            this.healthyTargets = results.filter((t) => !!t);
        });
    }
    /**
     * pick a healthy target randomly
     * if `enableHealthCheck` is `false`, this can not guarantee the returned target is healthy
     * @return {string} proxy target
     */
    pick() {
        const maxIndex = this.healthyTargets.length;
        if (maxIndex === 0) {
            throw Error('No healthy target, please check the proxy server!');
        }
        const randomIndex = MATH.floor(MATH.random() * maxIndex);
        const randomTarget = this.healthyTargets[randomIndex];
        logger.log(`Picked target: '${randomTarget}'`);
        return randomTarget;
    }
}
exports.default = ProxyLoadBalancer;
//# sourceMappingURL=proxy-load-balancer.js.map