import log = require('fary-log');
import http = require('http');
import Koa = require('koa');
import koaHttpProxy = require('koa-better-http-proxy');
import koaCompress = require('koa-compress');
import koaLogger = require('koa-logger');
import errorMiddleware = require('./middlewares/error');
import proxyMiddleware = require('./middlewares/proxy');
import checkProxy = require('./utils/check-proxy');
import ENV = require('./utils/env');
import serverConfig = require('./utils/server-config');

import { IAppOptions, IProxyMiddlewareOptions, IServerConfig } from './types';

const { IS_DEV } = ENV;
const logger = log('fary:app');

/**
 * createApp
 * @param {Object} options - options
 * @param {Function} onStart - callback after server start
 * @return {Koa} app - serverApp
 */
// tslint:disable-next-line:no-empty
function createApp({
  port = serverConfig.value.port,
  middlewares = [],
  skipListen = false,
  proxy = serverConfig.value.proxy,
  preProxy,
  postProxy,
// tslint:disable-next-line:align
}: IAppOptions = {}, onStart = () => {}) {
  checkProxy(proxy);

  logger.log(`Debug mode:`, IS_DEV);
  logger.log(`Proxy mapping:`, proxy);

  const app: Koa = new Koa();

  // trust proxy
  app.proxy = true;

  const allMiddlewares = [
    // request/response log
    koaLogger(),
    // compress the response
    koaCompress(),
    // error handler
    errorMiddleware(),
    ...middlewares,
    // proxy handler
    proxyMiddleware({
      postProxy,
      preProxy,
    }),
  ];

  allMiddlewares.forEach((middleware: Koa.Middleware) => {
    if (!middleware) {
      return;
    }
    app.use(middleware);
  });

  app.on('error', (err: Error) => {
    logger.error(`Server error: \n ${err.stack}`);
  });

  if (!skipListen) {
    app.listen(port, () => {
      logger.info(`Koa server is listening at ${port}`);
      onStart();
    });
  }

  process.on('SIGINT', () => {
    process.exit();
  });

  process.on('unhandledRejection', (err: Error) => {
    throw err;
  });

  return app;
}

// tslint:disable-next-line:no-namespace
declare namespace createApp {}

export = createApp;
